import csv
import pandas as pd
import numpy as np
from sklearn.model_selection import train_test_split
from sklearn.linear_model import LinearRegression, SGDClassifier, RidgeClassifier
from sklearn import svm
from sklearn.metrics import (
    matthews_corrcoef,
    f1_score,
    accuracy_score,
    roc_auc_score,
    precision_score,
    recall_score,
    balanced_accuracy_score
)
from sklearn.utils import resample
from typing import Iterable, List

from run_adni_with_dnabert import get_gene_location, generate_tsv_for_window_ADNI

RANDOM_SEED = 42
EVAL_FUNC_LIST = [
        accuracy_score,
        balanced_accuracy_score,
        precision_score,
        recall_score,
        f1_score,
        roc_auc_score,
        matthews_corrcoef,
    ]
HEADERS_LIST = [
                "SNP location within gene",
                "Accuracy",
                "Balanced accuracy",
                "Precision",
                "Recall",
                "F1 score",
                "ROC AUC",
                "MCC",
            ]
SVM_KERNEL = "rbf"


def run_svm(X_train, y_train, svm_kernel: str = "linear"): 
    np.random.seed(RANDOM_SEED)
    if svm_kernel == "linear":
        model = SGDClassifier()
    else:
        model = svm.SVC(kernel=svm_kernel)
    model.fit(X_train, y_train.values.ravel())
    return model


def evaluate(model, X_test, y_test, eval_func_list=EVAL_FUNC_LIST) -> List:
    y_pred = model.predict(X_test)
    y_decision = model.decision_function(X_test)

    results_list = []
    for func in eval_func_list:
        if func == roc_auc_score:
            result = func(y_test, y_decision)
        else:
            result = func(y_test, y_pred)
        results_list.append(round(result, 2))

    return results_list


def split_sequences_into_individual_cells(
    input_filepath: str, output_filepath: str
) -> None:
    """Enables visualization of SNPS within DNA sequences in Excel by splitting
    each nucleotide into its own column in a CSV."""
    with open(output_filepath, "w+") as output_file:
        writer = csv.writer(output_file)

        with open(input_filepath, newline="") as input_file:
            csv_reader = csv.reader(input_file, delimiter="\t")
            for row in csv_reader:
                seq_list = list(row[0])
                seq_list.append(row[1])  # add case/control label as last column
                writer.writerow(seq_list)


def run_linear_model_on_col(
    col: pd.core.series.Series,
    labels: pd.core.series.Series,
    svm_kernel: str,
    test_fraction: float = 0.1,
) -> List:
    """svm_kernel options are the options linear and those provided by svm.SVC"""
    # one-hot encode column
    X = pd.get_dummies(col)

    # train/test split
    X_train, X_test, y_train, y_test = train_test_split(
        X, labels, test_size=test_fraction, random_state=RANDOM_SEED
    )

    # run SVM model
    model = run_svm(X_train, y_train, svm_kernel)

    # Evaluation
    return evaluate(model, X_test, y_test)


def run_model_on_cols(
    df: pd.DataFrame,
    col_indices: Iterable[int],
    labels: pd.core.series.Series,
    model_func,
    output_csv_path: str,
) -> None:
    with open(output_csv_path, "w") as csv_file:
        writer = csv.writer(csv_file, delimiter=",")
        writer.writerow(HEADERS_LIST)
        for i in col_indices:
            col = df.iloc[:, i]
            results_list = model_func(col, labels, svm_kernel=SVM_KERNEL)
            row = [i]
            row.extend(results_list)
            writer.writerow(row)


def run_linear_model_on_col_with_downsample(
    col: pd.core.series.Series,
    labels: pd.core.series.Series,
    svm_kernel: str,
    test_fraction: float = 0.1,
):
    # concatenate data and labels together
    df = pd.concat([col, labels], axis=1)

    # Separate majority and minority classes
    df_majority = df[df.iloc[:, 1] == 1]
    df_minority = df[df.iloc[:, 1] == 0]

    # Downsample majority class
    df_majority_downsampled = resample(
        df_majority,
        replace=False,  # sample without replacement
        n_samples=df_minority.shape[0],
    )  # to match minority class

    # Combine minority class with downsampled majority class
    df_downsampled = pd.concat([df_majority_downsampled, df_minority])

    # one-hot encode column
    X = pd.get_dummies(df_downsampled.iloc[:, :-1])
    y = df_downsampled.iloc[:, -1:]

    # train/test split
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=test_fraction, random_state=RANDOM_SEED, stratify=y
    )

    # run SVM model
    model = run_svm(X_train, y_train, svm_kernel)

    # Evaluation
    return evaluate(model, X_test, y_test)


def run_linear_model_on_all_cols(
    snp_df: pd.DataFrame, labels: pd.core.series.Series, svm_kernel: str, test_fraction: float = 0.1 
) -> List:
    # one-hot encode column
    X = pd.get_dummies(snp_df)

    # train/test split
    X_train, X_test, y_train, y_test = train_test_split(
        X, labels, test_size=test_fraction, random_state=RANDOM_SEED
    )

    # run SVM model
    model = run_svm(X_train, y_train, svm_kernel)

    # Evaluation
    return evaluate(model, X_test, y_test)


def run_model_on_df(
    df: pd.DataFrame, labels: pd.core.series.Series, model_func, output_csv_path: str
) -> None:
    with open(output_csv_path, "w") as csv_file:
        writer = csv.writer(csv_file, delimiter=",")
        writer.writerow(HEADERS_LIST[1:])
        results_list = model_func(df, labels, svm_kernel=SVM_KERNEL)
        writer.writerow(results_list)


def run_linear_model_on_all_cols_with_downsample(
    snp_df: pd.DataFrame, labels: pd.core.series.Series, svm_kernel: str, test_fraction: float = 0.1 
) -> List:
    # concatenate data and labels together
    df = pd.concat([snp_df, labels], axis=1)

    # Separate majority and minority classes
    df_majority = df[df.iloc[:, -1] == 1]
    df_minority = df[df.iloc[:, -1] == 0]

    # Downsample majority class
    df_majority_downsampled = resample(
        df_majority,
        replace=False,  # sample without replacement
        n_samples=df_minority.shape[0],
    )  # to match minority class

    # Combine minority class with downsampled majority class
    df_downsampled = pd.concat([df_majority_downsampled, df_minority])

    # one-hot encode column
    X = pd.get_dummies(df_downsampled.iloc[:, :-1])
    y = df_downsampled.iloc[:, -1:]

    # train/test split
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=test_fraction, random_state=RANDOM_SEED, stratify=y
    )

    # run SVM model
    model = run_svm(X_train, y_train, svm_kernel)

    # Evaluation
    return evaluate(model, X_test, y_test)


def analysis_pipeline(gene_name: str) -> None:
    # output files
    svm_results_by_col_filename = "/home/user/svm_results/{}_svm_results_by_col.csv".format(
        gene_name
    )
    svm_results_by_col_downsample_filename = "/home/user/svm_results/{}_svm_results_by_col_downsample.csv".format(
        gene_name
    )
    svm_results_all_cols_filename = "/home/user/svm_results/{}_svm_results_all_cols.csv".format(
        gene_name
    )
    svm_results_all_cols_downsample_filename = "/home/user/svm_results/{}_svm_results_all_cols_downsample.csv".format(
        gene_name
    )

    # prepare data for analysis
    sequence_column_filename = _generate_sequence_column_filename(gene_name)
    df = pd.read_csv(sequence_column_filename, header=None)
    seq_df = df.iloc[:, :-1]
    labels = df.iloc[:, -1:]

    # SNP discovery 
    snp_cols = seq_df.columns[seq_df.nunique() > 1]
    print(snp_cols)
    print("Number of columns with mutations:", len(snp_cols))

    # run models
    run_model_on_cols(
        df, snp_cols, labels, run_linear_model_on_col, svm_results_by_col_filename
    )
    run_model_on_cols(df, snp_cols, labels, run_linear_model_on_col_with_downsample, svm_results_by_col_downsample_filename)

    snp_df = df[snp_cols]
    run_model_on_df(snp_df, labels, run_linear_model_on_all_cols, svm_results_all_cols_filename)
    run_model_on_df(snp_df, labels, run_linear_model_on_all_cols_with_downsample, svm_results_all_cols_downsample_filename)


def _generate_sequence_column_filename(gene_name) -> str:
    return "/home/user/DNABERT_model/data/{}_seq_cols.tsv".format(
        gene_name
    )


def generate_data_files(gene_name: str) -> None:
    # input files
    gene_annot_filename = "/home/user/Alzheimers-genetics-NNs/gene_annot.csv"
    case_control_filepath = "/home/user/ADNI_case_control.txt"
    fasta_directory = "/home/user/fasta1/"

    # intermediate steps files
    sequence_filename = "/home/user/DNABERT_model/data/{}_seqs.tsv".format(gene_name)
    sequence_column_filename = _generate_sequence_column_filename(gene_name)

    # get DNA sequences from ADNI dataset
    gene_loc = get_gene_location(gene_name, gene_annot_filename)
    generate_tsv_for_window_ADNI(
        gene_loc, case_control_filepath, sequence_filename, fasta_directory, False
    )
    split_sequences_into_individual_cells(sequence_filename, sequence_column_filename)


def main():
    np.random.seed(RANDOM_SEED)

    analysis_pipeline("APOE")

    # generate_data_files("SIGLEC11")
    # analysis_pipeline("SIGLEC11")

    # for gene in ["EXOC3L2", "TOMM40"]:
    # for gene in ["SIGLEC11", "APOE"]:
        # generate_data_files(gene)
        # analysis_pipeline(gene)


if __name__ == "__main__":
    main()
